use std::{
    collections::{HashMap, VecDeque},
    fs::File,
    io::Write,
    path::Path,
    time::Duration,
};

use chrono::{DateTime, TimeZone, Utc};
use clap::Parser;
use futures::TryStreamExt;
use indicatif::ProgressBar;
use qp_trie::Trie;
use rand::{
    distributions::{Distribution, WeightedIndex},
    Rng,
};
use regex::Regex;
use reqwest::{header, ClientBuilder};
use serde::{Deserialize, Serialize};
use smallvec::SmallVec;
use sqlx::postgres::PgPoolOptions;
use tracing::{debug, error, info, level_filters::LevelFilter};
use tracing_indicatif::IndicatifLayer;
use tracing_subscriber::{layer::SubscriberExt, util::SubscriberInitExt, EnvFilter};

#[derive(Clone, Copy, Debug, Hash, PartialEq, Eq, PartialOrd, Ord)]
enum MarkovItem {
    Start,
    Stop,
    Token(u32),
}

impl MarkovItem {
    fn index(self) -> u32 {
        match self {
            MarkovItem::Start => 0,
            MarkovItem::Stop => 1,
            MarkovItem::Token(idx) => idx + 2,
        }
    }

    fn from_u32(ind: u32) -> Self {
        match ind {
            0 => MarkovItem::Start,
            1 => MarkovItem::Stop,
            x => MarkovItem::Token(x.checked_sub(2).unwrap()),
        }
    }
}

const CONTEXT_SIZE_MAX: usize = 8;
const CONTEXT_SIZE_BYTES: usize = CONTEXT_SIZE_MAX * std::mem::size_of::<u32>();

type PrefixBytes = SmallVec<[u8; CONTEXT_SIZE_BYTES]>;

trait IntoBytes {
    fn into_bytes(self) -> PrefixBytes;
}

impl<'a> IntoBytes for &'a [MarkovItem] {
    fn into_bytes(self) -> PrefixBytes {
        self.iter()
            .copied()
            .map(MarkovItem::index)
            .flat_map(u32::to_be_bytes)
            .collect::<PrefixBytes>()
    }
}

#[derive(Debug, Default)]
struct Tokenizer {}

impl Tokenizer {
    fn tokenize<'a>(&self, input: &'a str) -> impl Iterator<Item = &'a str> {
        input.split_whitespace()
    }
}

#[derive(Debug, Serialize, Deserialize)]
struct State {
    model: Trie<PrefixBytes, f32>,
    mappings: HashMap<u32, String>,
    reverse_mappings: HashMap<String, u32>,
    fetched_up_to: Option<DateTime<Utc>>,
}

#[derive(Debug)]
struct Model<'a> {
    state: State,
    config: &'a Config,
    tokenizer: Tokenizer,
}

impl<'a> Model<'a> {
    fn new(config: &'a Config) -> Self {
        let mut model = Trie::new();
        model.insert(
            [MarkovItem::Start, MarkovItem::Stop].as_ref().into_bytes(),
            1.0,
        );

        Self::from_existing(
            config,
            State {
                model,
                mappings: HashMap::new(),
                reverse_mappings: HashMap::new(),
                fetched_up_to: None,
            },
        )
    }

    fn from_existing(config: &'a Config, state: State) -> Self {
        Self {
            state,
            config,
            tokenizer: Default::default(),
        }
    }

    #[tracing::instrument]
    fn generate(&self) -> String {
        let mut rand = rand::thread_rng();

        let mut attempts = 0;

        loop {
            let mut current = MarkovItem::Start;
            let mut tokens = Vec::new();
            let mut history = VecDeque::new();
            history.push_front(current);

            loop {
                while history.len() > self.config.history_len() {
                    history.pop_front();
                }

                if matches!(current, MarkovItem::Stop) {
                    break;
                }

                let mut options = HashMap::<u32, f32>::new();
                let prefix = history.make_contiguous().as_ref().into_bytes();
                let mut prefix_slice = prefix.as_ref();

                while !prefix_slice.is_empty() {
                    let item_size = std::mem::size_of::<u32>();
                    let prefix_len_bytes = prefix_slice.len();
                    let prefix_len_tokens = prefix_len_bytes / item_size;

                    // Mix in all bigrams if we're at the beginning
                    if matches!(current, MarkovItem::Start) {
                        options.extend(
                            self.state
                                .model
                                .iter()
                                .map(|(k, _)| k)
                                .filter(|k| k.len() == 2 * item_size)
                                .filter(|k| {
                                    !matches!(
                                        MarkovItem::from_u32(u32::from_be_bytes(
                                            k[item_size..].try_into().unwrap()
                                        )),
                                        MarkovItem::Stop
                                    )
                                })
                                .map(|item| {
                                    (
                                        u32::from_be_bytes(item[..item_size].try_into().unwrap()),
                                        rand.gen_range(0.5..2.0),
                                    )
                                }),
                        );
                    }

                    let prefixes = self
                        .state
                        .model
                        .iter_prefix(prefix_slice)
                        .filter(|(k, _)| k.len() >= prefix_len_bytes + item_size);

                    for (p, t) in prefixes {
                        let value = options
                            .entry(u32::from_be_bytes(
                                p[prefix_len_bytes..][..item_size].try_into().unwrap(),
                            ))
                            .or_insert(0f32);

                        *value += t * self.config.context_weights[prefix_len_tokens - 1];

                        /*
                        if matches!(current, MarkovItem::Start) {
                            *value = value.powf(0.75);
                        }
                        */
                    }

                    prefix_slice = &prefix_slice[item_size..];
                }

                let entries = options.into_iter().collect::<Vec<_>>();
                let weights = WeightedIndex::new(entries.iter().map(|v| v.1))
                    .expect("Weighted index failed to create");

                current = MarkovItem::from_u32(entries[weights.sample(&mut rand)].0);
                history.push_back(current);

                if let MarkovItem::Token(ti) = current {
                    tokens.push(
                        self.state
                            .mappings
                            .get(&ti)
                            .expect("Item should be present in the mappings map")
                            .clone(),
                    )
                }
            }

            if (tokens.is_empty() || tokens.join("").trim().is_empty())
                && attempts < self.config.attempts
            {
                attempts += 1;
                continue;
            }

            break tokens.join(" ").replace('@', "@\u{200b}");
        }
    }

    fn insert(&mut self, items: &[MarkovItem]) {
        let mut item_value = 100.0;

        if matches!(items[0], MarkovItem::Start)
            && matches!(items[1], MarkovItem::Token(t) if self.state.mappings.get(&t).is_some_and(|t| t.starts_with('@')))
        {
            item_value *= 1.0 - self.config.mention_penalty;
        }

        if matches!(items.last(), Some(MarkovItem::Stop)) {
            item_value *= self.config.stop_word_modifier;
        }

        if let Some(item) = self.state.model.get_mut(&items.into_bytes()) {
            *item += item_value;
        } else {
            self.state.model.insert(items.into_bytes(), item_value);
        }
    }

    fn convert_token(&mut self, tok: &str) -> MarkovItem {
        match self.state.reverse_mappings.get(tok) {
            Some(i) => MarkovItem::Token(*i),
            None => {
                let new_idx = self.state.reverse_mappings.len() as u32;
                self.state.reverse_mappings.insert(tok.to_owned(), new_idx);
                self.state.mappings.insert(new_idx, tok.to_owned());

                MarkovItem::Token(new_idx)
            }
        }
    }

    fn insert_tokens(&mut self, input: &str) {
        let input = input.trim();
        // Remove quotes
        let input = Regex::new(r"^> ?.+")
            .unwrap()
            .replace_all(input, "")
            .to_string();

        if input.contains("\n\n") {
            let paragraph_split = Regex::new("\n\n+").unwrap();
            paragraph_split
                .split(&input)
                .for_each(|p| self.insert_tokens(p));
            return;
        }

        if self
            .config
            .nasty_words
            .iter()
            .any(|w| input.to_lowercase().contains(w))
        {
            return;
        }
        let mut tokens = vec![MarkovItem::Start];
        tokens.extend(self.tokenizer.tokenize(&input).map(|t| {
            if t.starts_with("https") {
                self.convert_token(t)
            } else {
                self.convert_token(&t.to_lowercase())
            }
        }));
        for _ in 0..self.config.history_len() {
            tokens.push(MarkovItem::Stop);
        }

        for tok in tokens.windows(tokens.len().min(self.config.training_len()).max(2)) {
            self.insert(tok);
        }
    }
}

#[derive(Debug, Parser)]
#[command(version, about)]
struct Args {
    #[arg(long, default_value = "config.toml")]
    config_path: String,
    #[arg(long, default_value = "state.msg")]
    state_file: String,
}

#[derive(Debug, Deserialize)]
struct Config {
    #[serde(default)]
    learn_from_cws: bool,
    #[serde(default)]
    allowed_cws: Vec<String>,
    user_id: String,
    instance: String,
    period_seconds: u32,
    #[serde(default)]
    test_mode: bool,
    #[serde(default)]
    display_cw: Option<String>,
    context_weights: Vec<f32>,
    #[serde(default)]
    nasty_words: Vec<String>,
    #[serde(default)]
    attempts: usize,
    stop_word_modifier: f32,
    mention_penalty: f32,
}

impl Config {
    fn history_len(&self) -> usize {
        self.context_weights.len()
    }

    fn training_len(&self) -> usize {
        self.history_len() + 1
    }
}

#[tokio::main]
async fn main() {
    dotenvy::dotenv().ok();

    let indicatif_layer = IndicatifLayer::new();

    tracing_subscriber::registry()
        .with(
            EnvFilter::builder()
                .with_default_directive(LevelFilter::INFO.into())
                .from_env_lossy(),
        )
        .with(tracing_subscriber::fmt::layer().with_writer(indicatif_layer.get_stderr_writer()))
        .with(indicatif_layer)
        .init();

    let token = std::env::var("TOKEN").expect("token");

    let args = Args::parse();

    let config =
        toml::de::from_str::<Config>(&std::fs::read_to_string(&args.config_path).unwrap()).unwrap();

    if config.period_seconds < 1 {
        error!("Cannot have a period shorter than 1 second!");
        return;
    }

    if !(2..=CONTEXT_SIZE_MAX).contains(&config.training_len()) {
        error!(
            "Context size {} out of range {:?}",
            config.training_len(),
            2..=CONTEXT_SIZE_MAX
        );
        return;
    }

    let state_path = Path::new(&args.state_file);
    let progress = ProgressBar::new_spinner();

    let mut model = if state_path.is_file() {
        progress.set_message("Loading data...");
        let state =
            rmp_serde::decode::from_slice::<State>(&tokio::fs::read(state_path).await.unwrap())
                .unwrap();
        Model::from_existing(&config, state)
    } else {
        Model::new(&config)
    };

    let pool = PgPoolOptions::new()
        .max_connections(1)
        .connect(std::env::var("DATABASE_URL").as_deref().unwrap())
        .await
        .unwrap();

    let note_time_min = model
        .state
        .fetched_up_to
        .unwrap_or(Utc.with_ymd_and_hms(2010, 1, 1, 12, 0, 0).unwrap());

    progress.set_message("Fetching notes...");
    let text_stream = sqlx::query!(
        r#"SELECT text, "createdAt"
        FROM note
        WHERE "note"."userId" = $1 
            AND "note"."createdAt" > $2
            AND "note"."createdAt" < NOW() 
            AND "note"."visibility" IN ('public', 'home') 
            AND ($4 OR ("note"."cw" IS NULL OR LOWER("note"."cw") IN (SELECT UNNEST($3::VARCHAR[]))))"#,
        &config.user_id,
        note_time_min,
        &config.allowed_cws,
        config.learn_from_cws
    )
    .fetch(&pool);

    {
        let mut stream = progress.wrap_stream(text_stream);
        let mut cnt = 0;
        while let Ok(Some(item)) = stream.try_next().await.map_err(|e| {
            error!("Error fetching: {}", e);
            e
        }) {
            model.state.fetched_up_to = Some(item.createdAt);
            model.insert_tokens(item.text.as_deref().unwrap_or_default());
            progress.set_message(format!(
                "Items: {} | Trie size: {} | Current: {:?}",
                cnt,
                model.state.model.count(),
                item.text.unwrap_or_default().clone()
            ));
            cnt += 1;
            progress.tick();
            progress.set_length(cnt);
        }
    }

    tokio::spawn(async move {
        drop(pool);
    });

    info!("Keys: {}", model.state.model.count());

    progress.set_message("Saving data...");
    let encoded = rmp_serde::encode::to_vec(&model.state).unwrap();
    let file = File::create(state_path).unwrap();
    progress.wrap_write(file).write_all(&encoded).unwrap();

    let mut timer = tokio::time::interval(Duration::from_secs(config.period_seconds.into()));

    loop {
        timer.tick().await;

        let text = model.generate();
        info!("Generated: {:?}", text);

        if config.test_mode {
            continue;
        }

        let client = ClientBuilder::new().https_only(true).build().unwrap();
        match client
            .post(format!("https://{}/api/notes/create", &config.instance))
            .json(&serde_json::json!({
                "text": text,
                "poll": Option::<()>::None,
                "cw": config.display_cw.as_deref(),
                "visibility": "home"
            }))
            .header(header::AUTHORIZATION, &format!("Bearer {}", token))
            .send()
            .await
        {
            Err(e) => error!("Fetch error: {:#?}", e),
            Ok(r) => debug!("Response: {:#?}", r.json::<serde_json::Value>().await),
        }
    }
}
